/*******************************************************************
 Project : SUN Legacy Type5 keyboard -> USB converter.
 Author  : IIDA Tetsushi (hogeman@gmail.com)
 Date    : 2009-01-20
 License : GNU GPL v2

 based on other projects such as "HIDKeys", "Dulcimer", etc. 
 *******************************************************************/
/*
#define F_CPU 16000000UL
*/

#include <avr/io.h>
#include <avr/pgmspace.h> 
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <string.h>

#include "usbdrv/usbdrv.h"

/*----------------------------------------------------------*/
/* definition about USB keyboard 		 					*/
/* see : "Device Class Definition for HID v.1.11"			*/ 
/*----------------------------------------------------------*/
/* USB keyboard report descriptor (created by USB.org's tool) */
PROGMEM char usbHidReportDescriptor[64] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x81, 0x01,                    //   INPUT (Cnst,Ary,Abs)
    0x95, 0x05,                    //   REPORT_COUNT (5)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x05, 0x08,                    //   USAGE_PAGE (LEDs)
    0x19, 0x01,                    //   USAGE_MINIMUM (Num Lock)
    0x29, 0x05,                    //   USAGE_MAXIMUM (Kana)
    0x91, 0x02,                    //   OUTPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x03,                    //   REPORT_SIZE (3)
    0x91, 0x01,                    //   OUTPUT (Cnst,Ary,Abs)
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};

/* USB keyboard report buffer */
static struct REPORT_BUF_STRUCT {
	uchar modKey;
	uchar reserved;
	uchar keys[6];
} reportBuffer;

static uchar idleRate;    		/* in 4 ms units */
static uchar protocolVer = 1;	/*  0 = boot protocol, 1 = report protocol */
static uchar expectReport;		/* flag to indicate if we expect an USB-report */
static uchar keyDidChange;		/* key status was changed */


/*----------------------------------------------------------*/
/* for macro command 										*/
/*----------------------------------------------------------*/
enum keyMacroState {
	KM_NORMAL = 0,		/* normal operation */
	KM_RECORDING,		/* on recording		*/
	KM_RUNNING			/* on running		*/
};
static enum keyMacroState kmStat;		/* current operation mode */
static uchar kmIdx;						/* stSeq buffer selector (0..3) */

struct STROKE_SEQUENCE_BUF {
	uchar len;			/* effective data length */
	uchar keys[63];		/* key stroke buffer */
};
static struct STROKE_SEQUENCE_BUF rom_stSeq[4] EEMEM = { /* 256bytes(=minimum size of ATMega?8) */
	{0, {0}}, {0, {0}}, {0, {0}}, {0, {0}}};			/* initialize by zero */
static struct STROKE_SEQUENCE_BUF stSeqOnRec;			/* key stroke Recording buffer */
static struct STROKE_SEQUENCE_BUF stSeq[4];				/* key stroke buffer */


/*----------------------------------------------------------*/
/* definition about SUN Keyboard (Serial) 					*/
/*----------------------------------------------------------*/
#define UBRR_Value(b)       ((F_CPU-8*b)/(16*b))

/* SUN keyboard commands */
#define SKBDCMD_RESET       0x01
#define SKBDCMD_BELLON      0x02
#define SKBDCMD_BELLOFF     0x03
#define SKBDCMD_SETLED      0x0e


/* The led set commands require sending the SETLED byte then
 * a byte encoding which led's to have set.  Here are the bit
 * values, a bit set = led-on. */
#define SKBD_LED_NLOCK          0x01   /* Num-lock */
#define SKBD_LED_CMPOSE         0x02   /* Compose */
#define SKBD_LED_SCRLCK         0x04   /* Scroll-lock */
#define SKBD_LED_CLOCK          0x08   /* Caps-lock */
#define SKBD_LED_KANA           0x10   /* kana */

/* USB equivalents. order of bits are different from SUN's definition. */
#define USB_LED_NLOCK          	0x01   /* Num-lock */
#define USB_LED_CLOCK          	0x02   /* Caps-lock */
#define USB_LED_SCRLCK        	0x04   /* Scroll-lock */
#define USB_LED_CMPOSE        	0x08   /* Compose */
#define USB_LED_KANA          	0x10   /* kana */

/* Special state characters */
#define SKBD_RESET          0xff
#define SKBD_LYOUT          0xfe
#define SKBD_ALLUP          0x7f

/* Special Keys */
#define SKBD_HELP			0x76
#define SKBD_STOP			0x01
#define SKBD_AGAIN			0x03
#define SKBD_PROPS			0x19
#define SKBD_UNDO			0x1a
#define SKBD_FRONT			0x31
#define SKBD_COPY			0x33
#define SKBD_OPEN			0x48
#define SKBD_PASTE			0x49
#define SKBD_FIND			0x5f
#define SKBD_CUT			0x61
/* #define SKBD_LINEFEED		0x6F */ /* Japanese Keyboard uses this key as normal character */
#define SKBD_POWER			0x30


/* SUN keycode -> USB Usage ID mapping. (zero == invalid)  */
#define JAPANESE_KEYBOARD
#ifdef  JAPANESE_KEYBOARD
/* this is for Japanes keyboard. */
const prog_uchar sun_type5_usbmap[128] = {
   0,  0,  129,   0, 128,  58,  59,  67,	/* 0x00-0x07 */
  60,  68,  61,  69,  62, 136,  63,   0,	/* 0x08-0x0f */
  64,  65,  66, 226,  82,  72,  70,  71,	/* 0x10-0x17 */
  80,   0,   0,  81,  79,  41,  30,  31,	/* 0x18-0x1f */
  32,  33,  34,  35,  36,  37,  38,  39,	/* 0x20-0x27 */
  45,  46, 137,  42,  73, 127,  84,  85,	/* 0x28-0x2f */
   0,   0,  99,   0,  74,  43,  20,  26,	/* 0x30-0x37 */
   8,  21,  23,  28,  24,  12,  18,  19,	/* 0x38-0x3f */
  47,  48,  76, 101,  95,  96,  97,  86,	/* 0x40-0x47 */
   0,   0,  77,   0, 224,   4,  22,   7,	/* 0x48-0x4f */
   9,  10,  11,  13,  14,  15,  51,  52,	/* 0x50-0x57 */
  50,  40,  88,  92,  93,  94,  98,   0,	/* 0x58-0x5f */
  75,   0,  83, 225,  29,  27,   6,  25,	/* 0x60-0x67 */
   5,  17,  16,  54,  55,  56, 229, 135,	/* 0x68-0x6f */
  89,  90,  91, 139, 138,  53,   0,  57,	/* 0x70-0x77 */
 227,  44, 231,  78,   0,  87,   0,   0 	/* 0x78-0x7f */
};
#else
/* if you need US keyboard definition, try this. 
 * this is not tested because I've never owned US keyboard.
 * if you find a bug, please contact me.
 */
const prog_uchar sun_type5_usbmap[128] = {
   0,  0,  129,   0, 128,  58,  59,  67,	/* 0x00-0x07 */
  60,  68,  61,  69,  62, 230,  63,   0,	/* 0x08-0x0f */
  64,  65,  66, 226,  82,  72,  70,  71,	/* 0x10-0x17 */
  80,   0,   0,  81,  79,  41,  30,  31,	/* 0x18-0x1f */
  32,  33,  34,  35,  36,  37,  38,  39,	/* 0x20-0x27 */
  45,  46,  53,  42,  73, 127,  84,  85,	/* 0x28-0x2f */
   0,   0,  99,   0,  74,  43,  20,  26,	/* 0x30-0x37 */
   8,  21,  23,  28,  24,  12,  18,  19,	/* 0x38-0x3f */
  47,  48,  76, 101,  95,  96,  97,  86,	/* 0x40-0x47 */
   0,   0,  77,   0, 224,   4,  22,   7,	/* 0x48-0x4f */
   9,  10,  11,  13,  14,  15,  51,  52,	/* 0x50-0x57 */
  49,  40,  88,  92,  93,  94,  98,   0,	/* 0x58-0x5f */
  75,   0,  83, 225,  29,  27,   6,  25,	/* 0x60-0x67 */
   5,  17,  16,  54,  55,  56, 229,   0,	/* 0x68-0x6f */
  89,  90,  91,   0,   0,   0,   0,  57,	/* 0x70-0x77 */
 227,  44, 231,  78,   0,  87,   0,   0 	/* 0x78-0x7f */
};
#endif

/*----------------------------------------------------------*/
/* Function Prototypes					 					*/
/*----------------------------------------------------------*/
static void wait_ms(int);
static void hardwareInit(void);
static void uart_putchar(uchar);
static uchar sunCode2UsageId(uchar);

/*----------------------------------------------------------*/
/* functions from here					 					*/
/*----------------------------------------------------------*/

/**
 * Wait function. 
 */
static void wait_ms(int m)
{
	while(m--) _delay_ms(1);
}

/**
 * IO Port initialization.
 */
static void hardwareInit(void)
{
	/*
	  PB0    -> Power Indicator LED (for debug) 
	  PB1..4 -> ICSP
	  PB5..6 -> Ext.Crystal
     */
    PORTB = 0xff;   /* activate all pull-ups */
    DDRB = 0x01;    /* all pins input except LED */
	/*
	  PC0..5 -> (NOT USED) (LCD for debug)
	  PC6    -> RESET
	 */
    PORTC = 0xff;   /* activate all pull-ups */
    DDRC = 0x00;    /* all pins input */
    /*
	  PD0..1 -> TX/RX
	  PD2..3 -> D+/D-
	  PD4    -> (NOT USED)
	  PD5    -> USB PULLUP CTRL
	  PD6..7 -> (NOT USED)
	 */
    PORTD = 0xd3;   /* 1101 0011 bin: activate pull-ups except on USB lines */
    DDRD  = 0x20;   /* 0010 0000 bin: all pins input except USB reset */

	/* USB disconnect -> reconnect */
	usbDeviceDisconnect();
	wait_ms(50);   /* wait 50ms */	/* probably this is not required as following lines are vely slow */

	/* restore keystroke previously saved from EEPROM (slow) */
	eeprom_busy_wait();
	eeprom_read_block(&stSeq, &rom_stSeq, sizeof(stSeq));	/* read saved keystroke */

	usbDeviceConnect();

	usbInit();

    /* configure timer 0 for a rate of 16M/(256 * 256) = 244.14 Hz (4.096ms) */
    TCCR0B = 4;      /* timer 0 prescaler: 256 */
	
	/* serial settings */
	UBRR0H = UBRR_Value(1200) >> 8;
	UBRR0L = UBRR_Value(1200) &  0xFF;
	UCSR0B = _BV(RXEN0) | _BV(TXEN0) | _BV(RXCIE0);	
}

/**
 * SEND command to SUN Keyboard.
 */
static void uart_putchar(uchar c)
{
	loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = c;
}

/**
 * build a USB ReportBuffer from Sun Keyboard Code
 */ 
static uchar sunCode2UsageId(uchar rb)
{
	uchar cnt;
	uchar keyUp = rb & 0x80;		/* 0:Down, ~0:Up */

	/* Get USB Keycode */
	uchar usbKey = pgm_read_byte(&(sun_type5_usbmap[rb & 0x7f]));

	/* ignore N/A key */
	if (usbKey == 0) {
		return 0;			/* not changed */

	/* Check Modifier Keys */
	} else if ((0xe0 <= usbKey) && (usbKey <= 0xe7)) {
		if (keyUp) { 
			reportBuffer.modKey &=  ~(1 << (usbKey - 0xe0));
		} else {
			reportBuffer.modKey |=  (1 << (usbKey - 0xe0));
		}
		
	/* Check Normal Keys */	
	} else {
		/* parse keys */
		if (keyUp) {			/* when key release */
			/* draw ahead required? */ 
			for (cnt = 0; cnt < sizeof reportBuffer.keys; cnt++) {
				if (reportBuffer.keys[cnt] == usbKey) {
					reportBuffer.keys[cnt] = 0;
					break;
				}
			}
		} else {				/* when key press */
			for (cnt = 0; cnt < sizeof reportBuffer.keys; cnt++) {
				if (reportBuffer.keys[cnt] == 0) {
					reportBuffer.keys[cnt] = usbKey;
					break;
				}
			}
		}
	} 
	return 1;			/* key did change */
}

/**
 * RECV data from SUN Keyboard.
 */
ISR(USART_RX_vect)
{
    static uchar pwrOnFlag;  		/* to ignore "0xff 0x04" when power on. */

    uchar rb = UDR0;
	uchar keyUp = rb & 0x80;		/* 0:Down, ~0:Up */
	uchar vKey  = rb & 0x7f;

	sei();				/* for USB interrupt. see usbdrv.h */

	/* ignore any key on macro running */
	if (kmStat == KM_RUNNING) {
		return;
	}

	/* On Power ON ... */
	if (rb == SKBD_RESET) {
		pwrOnFlag = 1;	/* ignore following character */
		return;
	}
	if (pwrOnFlag) {
		pwrOnFlag = 0;
		return;			/* ignore keyboard type */
	}

	/* reset keyboard state.  */
	if (rb == SKBD_ALLUP) {
		memset((void *)&reportBuffer, 0, sizeof reportBuffer);
		keyDidChange = 1;

	/* Check "Start Recording" key */
	} else if (vKey == SKBD_HELP) {
		if (keyUp) {
			memset(&stSeqOnRec, 0, sizeof stSeqOnRec);
			kmStat = KM_RECORDING;
		}

	/* Check "Run" key */
	} else if ((vKey == SKBD_STOP) || (vKey == SKBD_PROPS) || (vKey == SKBD_FRONT) || (vKey == SKBD_OPEN)) {
		if (keyUp && kmStat == KM_NORMAL) {
			if (vKey == SKBD_STOP)
				kmIdx = 0; 
			else if (vKey == SKBD_PROPS)
				kmIdx = 1;
			else if (vKey == SKBD_FRONT)
				kmIdx = 2;
			else 
				kmIdx = 3;

			kmStat = KM_RUNNING;
			/* execute at main loop */
		}

	/* Check "Save" key */
	} else if ((vKey == SKBD_AGAIN) || (vKey == SKBD_UNDO) || (vKey == SKBD_COPY) || (vKey == SKBD_PASTE)) {
		if (keyUp && kmStat == KM_RECORDING) {
			kmStat = KM_NORMAL;
			uchar idx;
			if (vKey == SKBD_AGAIN)
				idx = 0; 
			else if (vKey == SKBD_UNDO)
				idx = 1;
			else if (vKey == SKBD_COPY)
				idx = 2;
			else 
				idx = 3;
			stSeq[idx] = stSeqOnRec;
			/* save to eeprom (very slow!) */
			eeprom_busy_wait();
			eeprom_write_block(&(stSeq[idx]), &(rom_stSeq[idx]), sizeof(stSeq[0]));
		}

	/* real key event */
	} else {
		/* record keystrokes */
		if (kmStat == KM_RECORDING) {
			if (stSeqOnRec.len < sizeof stSeqOnRec.keys) {
				stSeqOnRec.keys[stSeqOnRec.len] = rb;
				stSeqOnRec.len++;
			} else {
				kmStat = KM_NORMAL;
				uart_putchar(SKBDCMD_BELLON);
				memset(&stSeqOnRec, 0, sizeof stSeqOnRec);		
				uart_putchar(SKBDCMD_BELLOFF);		
			}
		}

		/* make usb keycode */
		if (sunCode2UsageId(rb)) {
			keyDidChange = 1;
		}
	}
}


/**
 * The write function is called when LEDs should be set. 
 */
uchar usbFunctionWrite(uchar *data, uchar len) {
	uchar cLED = 0;

    if (expectReport && (len == 1)) {
		if (data[0] & USB_LED_NLOCK) {
			cLED |= SKBD_LED_NLOCK;
		}
		if (data[0] & USB_LED_CLOCK) {
			cLED |= SKBD_LED_CLOCK;
		}
		if (data[0] & USB_LED_SCRLCK) {
			cLED |= SKBD_LED_SCRLCK;
		}
		if (data[0] & USB_LED_CMPOSE) {
			cLED |= SKBD_LED_CMPOSE;
		}
		if (data[0] & USB_LED_KANA) {
			cLED |= SKBD_LED_KANA;
		}
		uart_putchar(SKBDCMD_SETLED);
		uart_putchar(cLED);
	}
    expectReport = 0;
    return 1;
}

/**
 * This function is called whenever we receive a setup request via USB.
 */
usbMsgLen_t	usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (void *)data;

    usbMsgPtr = (uchar *)&reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_SET_REPORT){ 
            if (rq->wLength.word == 1) {
                /* We expect one byte reports */
                expectReport = 1;
                return USB_NO_MSG; /* Call usbFunctionWrite with data */
            }
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        } else if (rq->bRequest == USBRQ_HID_GET_PROTOCOL) {
		    if (rq->wValue.bytes[1] < 1) {
                protocolVer = rq->wValue.bytes[1];
            }
        } else if(rq->bRequest == USBRQ_HID_SET_PROTOCOL) {
            usbMsgPtr = &protocolVer;
            return 1;
        }

    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

/**
 * main function.
 */
int main(void)
{
	uchar   idleCounter = 0;
	uchar	sendCnt = 0;

	hardwareInit();

	wdt_enable(WDTO_1S);

	sei(); 			/* enable interrupts */

	for (;;) {		/* ensure loop within 50ms! */
		wdt_reset();
		usbPoll();

        if (TIFR0 & _BV(TOV0)) {		/* 4ms period expired */ 
            TIFR0 = _BV(TOV0);
            if(idleRate != 0){
                if(idleCounter > 0){
                    idleCounter--;
                }else{
                    idleCounter = idleRate;
                   	keyDidChange = 1;
                }
            }
        }

		if (kmStat == KM_RUNNING) {			/* macro mode */
			/* execute macro */
			if (sendCnt < stSeq[kmIdx].len) {
			 	if (usbInterruptIsReady()) {
					sunCode2UsageId(stSeq[kmIdx].keys[sendCnt]);
					usbSetInterrupt((uchar *)&reportBuffer, sizeof(reportBuffer)); // send
					sendCnt++;
				}
			} else {
				/* exit to normal mode */
				sendCnt = 0;
				kmStat = KM_NORMAL;
			}

		} else {
	        if(keyDidChange && usbInterruptIsReady()) {
		        keyDidChange = 0;
	            usbSetInterrupt((uchar *)&reportBuffer, sizeof(reportBuffer));
			}
		}
	}

	return(0);
}
